@@ -21,6 +21,7 @@
 
 import apt
 import atexit
+import distro_info
 import glob
 import logging
 import os
@@ -32,16 +33,6 @@ from subprocess import PIPE, Popen
 from .utils import get_arch
 
 from .DistUpgradeGettext import gettext as _
-try:
-    from janitor.plugincore.manager import PluginManager
-except ImportError:
-    # janitor is not available, so create a no-op plugin manager.
-    class PluginManager(object):
-        def __init__(self, *args, **kws):
-            pass
-
-        def get_plugins(self, *args, **kws):
-            return []
 
 
 class DistUpgradeQuirks(object):
@@ -58,9 +49,12 @@ class DistUpgradeQuirks(object):
         self.uname = Popen(["uname", "-r"], stdout=PIPE,
                            universal_newlines=True).communicate()[0].strip()
         self.arch = get_arch()
-        self.plugin_manager = PluginManager(self.controller, ["./plugins"])
+        self.extra_snap_space = 0
         self._poke = None
         self._snapstore_reachable = False
+        self._snap_list = None
+        self._from_version = None
+        self._to_version = None
 
     # the quirk function have the name:
     #  $Name (e.g. PostUpgrade)
@@ -87,15 +81,6 @@ class DistUpgradeQuirks(object):
             return
         to_release = self.config.get("Sources", "To")
         from_release = self.config.get("Sources", "From")
-        # first check for matching plugins
-        for condition in [
-                quirksName,
-                "%s%s" % (to_release, quirksName),
-                "from_%s%s" % (from_release, quirksName)
-        ]:
-            for plugin in self.plugin_manager.get_plugins(condition):
-                logging.debug("running quirks plugin %s" % plugin)
-                plugin.do_cleanup_cruft()
 
         # run the handler that is common to all dists
         funcname = "%s" % quirksName
@@ -124,10 +109,11 @@ class DistUpgradeQuirks(object):
         logging.debug("running Quirks.PreCacheOpen")
 
     # individual quirks handler that run *after* the cache is opened
-    def cosmicPostInitialUpdate(self):
+    def focalPostInitialUpdate(self):
         # PreCacheOpen would be better but controller.abort fails terribly
         """ run after the apt cache is opened the first time """
-        logging.debug("running Quirks.cosmicPostInitialUpdate")
+        logging.debug("running Quirks.focalPostInitialUpdate")
+        self._get_from_and_to_version()
         self._test_and_fail_on_i386()
         cache = self.controller.cache
         if 'ubuntu-desktop' not in cache or \
@@ -137,9 +123,14 @@ class DistUpgradeQuirks(object):
         if cache['ubuntu-desktop'].is_installed and \
                 cache['snapd'].is_installed:
             self._checkStoreConnectivity()
+        # If the snap store is accessible, at the same time calculate the
+        # extra size needed by to-be-installed snaps.  This also prepares
+        # the snaps-to-install list for the actual upgrade.
+        if self._snapstore_reachable:
+            self._calculateSnapSizeRequirements()
 
-    def cosmicPostUpgrade(self):
-        logging.debug("running Quirks.cosmicPostUpgrade")
+    def focalPostUpgrade(self):
+        logging.debug("running Quirks.focalPostUpgrade")
         cache = self.controller.cache
         if 'ubuntu-desktop' not in cache or \
                 'snapd' not in cache:
@@ -147,8 +138,8 @@ class DistUpgradeQuirks(object):
             return
         if cache['ubuntu-desktop'].is_installed and \
                 cache['snapd'].is_installed and \
-                self._snapstore_reachable:
-            self._replaceDebsWithSnaps()
+                self._snap_list:
+            self._replaceDebsAndSnaps()
 
     # individual quirks handler when the dpkg run is finished ---------
     def PostCleanup(self):
@@ -172,6 +163,7 @@ class DistUpgradeQuirks(object):
         """ run after calculating the dist-upgrade """
         logging.debug("running Quirks.PostDistUpgradeCache")
         self._install_linux_metapackage()
+        self._install_python_is_python2()
 
     # helpers
     def _get_pci_ids(self):
@@ -187,6 +179,24 @@ class DistUpgradeQuirks(object):
                 lspci.add(line.split()[2])
         return lspci
 
+    def _get_from_and_to_version(self):
+        di = distro_info.UbuntuDistroInfo()
+        try:
+            self._from_version = \
+                di.version('%s' % self.controller.fromDist).split()[0]
+            self._to_version = \
+                di.version('%s' % self.controller.toDist).split()[0]
+        # Ubuntu 18.04's python3-distro-info does not have version
+        except AttributeError:
+            self._from_version = next(
+                (r.version for r in di.get_all("object")
+                 if r.series == self.controller.fromDist),
+                self.controller.fromDist).split()[0]
+            self._to_version = next(
+                (r.version for r in di.get_all("object")
+                 if r.series == self.controller.toDist),
+                self.controller.toDist).split()[0]
+
     def _test_and_warn_for_unity_3d_support(self):
         UNITY_SUPPORT_TEST = "/usr/lib/nux/unity_support_test"
         if (not os.path.exists(UNITY_SUPPORT_TEST) or
@@ -288,15 +298,20 @@ class DistUpgradeQuirks(object):
     def _test_and_fail_on_i386(self):
         """
         Test and fail if the package architecture is i386 as we
-        may drop support for this architecture.
+        have dropped support for this architecture.
         """
+        if self._from_version == '18.04':
+            updates_end = 'April 2023'
+        elif self._from_version == '19.04':
+            updates_end = 'January 2020'
         # check on i386 only
         if self.arch == "i386":
-            logging.error("package arch is i386")
-            summary = _("apt architecture is i386")
-            msg = _("Upgrades from 18.04 on the i386 "
-                    "architecture are not supported "
-                    "at this time.")
+            logging.error("apt architecture is i386")
+            summary = _("Sorry, no more upgrades for this system")
+            msg = _("There will not be any further Ubuntu releases "
+                    "for this system's 'i386' architecture.\n\n"
+                    "Updates for Ubuntu %s will continue until %s." %
+                    (self._from_version, updates_end))
             self._view.error(summary, msg)
             self.controller.abort()
 
@@ -434,13 +449,37 @@ class DistUpgradeQuirks(object):
         # can't connect
         elif re.search("^ \*.*unreachable", connected[0], re.MULTILINE):
             logging.error("No snap store connectivity")
-            res = self._view.askYesNoQuestion(
-                _("Connection to Snap Store failed"),
-                _("Your system does not have a connection to the Snap "
-                  "Store. For the best upgrade experience make sure "
-                  "that your system can connect to api.snapcraft.io.\n"
-                  "Do you still want to continue with the upgrade?")
-            )
+            old_lxd_deb_installed = False
+            cache = self.controller.cache
+            if 'lxd' in cache:
+                # epoch 1 is the transitional deb
+                if cache['lxd'].is_installed and not \
+                        cache['lxd'].candidate.version.startswith("1:"):
+                    logging.error("lxd is installed")
+                    old_lxd_deb_installed = True
+            if old_lxd_deb_installed:
+                summary = _("Connection to the Snap Store failed")
+                msg = _("You have the package lxd installed but your "
+                        "system is unable to reach the Snap Store. "
+                        "lxd is now provided via a snap and the release "
+                        "upgrade will fail if snapd is not functional. "
+                        "Please make sure you're connected to the "
+                        "Internet and update any firewall or proxy "
+                        "settings as needed so that you can reach "
+                        "api.snapcraft.io. If you are an enterprise "
+                        "with a firewall setup you may want to configure "
+                        "a Snap Store proxy."
+                        )
+                self._view.error(summary, msg)
+                self.controller.abort()
+            else:
+                res = self._view.askYesNoQuestion(
+                    _("Connection to Snap Store failed"),
+                    _("Your system does not have a connection to the Snap "
+                      "Store. For the best upgrade experience make sure "
+                      "that your system can connect to api.snapcraft.io.\n"
+                      "Do you still want to continue with the upgrade?")
+                )
         # debug command not available
         elif 'error: unknown command' in connected[1]:
             logging.error("snap debug command not available")
@@ -458,44 +497,78 @@ class DistUpgradeQuirks(object):
         if not res:
             self.controller.abort()
 
-    def _replaceDebsWithSnaps(self):
+    def _calculateSnapSizeRequirements(self):
+        import json
+        import urllib.request
+        from urllib.error import URLError
+
+        # first fetch the list of snap-deb replacements that will be needed
+        # and store them for future reference, along with other data we'll
+        # need in the process
+        self._prepare_snap_replacement_data()
+        # now perform direct API calls to the store, requesting size
+        # information for each of the snaps needing installation
+        self._view.updateStatus(_("Calculating snap size requirements"))
+        for snap, snap_object in self._snap_list.items():
+            if snap_object['command'] != 'install':
+                continue
+            action = {
+                "instance-key": "upgrade-size-check",
+                "action": "download",
+                "snap-id": snap_object['snap-id'],
+                "channel": snap_object['channel'],
+            }
+            data = {
+                "context": [],
+                "actions": [action],
+            }
+            req = urllib.request.Request(
+                url='https://api.snapcraft.io/v2/snaps/refresh',
+                data=bytes(json.dumps(data), encoding='utf-8'))
+            req.add_header('Snap-Device-Series', '16')
+            req.add_header('Content-type', 'application/json')
+            req.add_header('Snap-Device-Architecture', self.arch)
+            try:
+                response = urllib.request.urlopen(req).read()
+                info = json.loads(response)
+                size = int(info['results'][0]['snap']['download']['size'])
+            except (KeyError, URLError, ValueError):
+                logging.debug("Failed fetching size of snap %s" % snap)
+                continue
+            self.extra_snap_space += size
+
+    def _replaceDebsAndSnaps(self):
         """ install a snap and mark its corresponding package for removal """
-        # gtk-common-themes isn't a package name but is this risky?
-        snaps = ['gnome-3-26-1604', 'gtk-common-themes', 'gnome-calculator',
-                 'gnome-characters', 'gnome-logs', 'gnome-system-monitor']
-        self._view.updateStatus(_("Checking for installed snaps"))
-        for snap in snaps:
-            # check to see if the snap is already installed
-            snap_info = subprocess.Popen(["snap", "info", snap],
-                                         universal_newlines=True,
-                                         stdout=subprocess.PIPE).communicate()
-            self._view.processEvents()
-            if re.search("^installed: ", snap_info[0], re.MULTILINE):
-                logging.debug("Snap %s is installed" % snap)
-                # its not tracking the release channel so don't refresh
-                if not re.search("^tracking:.*ubuntu-18.04", snap_info[0],
-                                 re.MULTILINE):
-                    logging.debug("Snap %s is not tracking the release channel"
-                                  % snap)
-                    continue
-                command = 'refresh'
+        self._view.updateStatus(_("Processing snap replacements"))
+        # _snap_list should be populated by the earlier
+        # _calculateSnapSizeRequirements call.
+        for snap, snap_object in self._snap_list.items():
+            command = snap_object['command']
+            if command == 'refresh':
+                self._view.updateStatus(_("refreshing snap %s" % snap))
+                popenargs = ["snap", command,
+                             "--channel", snap_object['channel'], snap]
+            elif command == 'remove':
+                self._view.updateStatus(_("removing snap %s" % snap))
+                popenargs = ["snap", command, snap]
             else:
-                command = 'install'
+                self._view.updateStatus(_("installing snap %s" % snap))
+                popenargs = ["snap", command,
+                             "--channel", snap_object['channel'], snap]
             try:
-                self._view.updateStatus(_("%sing snap %s" % (command, snap)))
                 self._view.processEvents()
-                proc = subprocess.run(["snap", command, "--channel",
-                                       "stable/ubuntu-18.10", snap],
-                                      stdout=subprocess.PIPE,
-                                      check=True)
+                proc = subprocess.run(
+                    popenargs,
+                    stdout=subprocess.PIPE,
+                    check=True)
                 self._view.processEvents()
             except subprocess.CalledProcessError:
                 logging.debug("%s of snap %s failed" % (command, snap))
                 continue
             if proc.returncode == 0:
                 logging.debug("%s of snap %s succeeded" % (command, snap))
-            if command == 'install':
-                self.controller.forced_obsoletes.append(snap)
+            if command == 'install' and snap_object['deb']:
+                self.controller.forced_obsoletes.append(snap_object['deb'])
 
     def _checkPae(self):
         " check PAE in /proc/cpuinfo "
@@ -761,6 +834,20 @@ class DistUpgradeQuirks(object):
             reason = "linux metapackage may have been accidentally uninstalled"
             cache.mark_install(linux_metapackage, reason)
 
+    def _install_python_is_python2(self):
+        """
+        Ensure the python-is-python2 is installed if python-minimal
+        was installed.
+        """
+        old = 'python-minimal'
+        new = 'python-is-python2'
+        cache = self.controller.cache
+        if old in cache and cache[old].is_installed:
+            logging.info("installing %s because %s was installed" % (new, old))
+            reason = "%s was installed on the system" % old
+            if not cache.mark_install(new, reason):
+                logging.info("failed to install %s" % new)
+
     def ensure_recommends_are_installed_on_desktops(self):
         """ ensure that on a desktop install recommends are installed
             (LP: #759262)
@@ -771,3 +858,116 @@ class DistUpgradeQuirks(object):
                 msg += " enabling it just for the upgrade"
                 logging.warning(msg)
                 apt.apt_pkg.config.set("Apt::Install-Recommends", "1")
+
+    def _prepare_snap_replacement_data(self):
+        """ Helper function fetching all required info for the deb-to-snap
+            migration: version strings for upgrade (from and to) and the list
+            of snaps (with actions).
+        """
+        import json
+        self._snap_list = {}
+        from_channel = "stable/ubuntu-%s" % self._from_version
+        to_channel = "stable/ubuntu-%s" % self._to_version
+        seeded_snaps = {}
+        unseeded_snaps = {}
+
+        try:
+            current_path = os.path.dirname(os.path.abspath(__file__))
+            d2s_file = open(current_path + '/deb2snap.json', 'r')
+            d2s = json.load(d2s_file)
+            d2s_file.close()
+
+            for snap in d2s["seeded"]:
+                seed = d2s["seeded"][snap]
+                deb = seed.get("deb", None)
+                from_chan = seed.get("from_channel", from_channel)
+                to_chan = seed.get("to_channel", to_channel)
+                seeded_snaps[snap] = (deb, from_chan, to_chan)
+
+            for snap in d2s["unseeded"]:
+                unseed = d2s["unseeded"][snap]
+                deb = unseed.get("deb", None)
+                from_chan = unseed.get("from_channel", from_channel)
+                unseeded_snaps[snap] = (deb, from_chan)
+        except Exception as e:
+            logging.warning("error reading deb2snap.json file (%s)" % e)
+
+        self._view.updateStatus(_("Checking for installed snaps"))
+        for snap, (deb, from_channel, to_channel) in seeded_snaps.items():
+            snap_object = {}
+            # check to see if the snap is already installed
+            snap_info = subprocess.Popen(["snap", "info", snap],
+                                         universal_newlines=True,
+                                         stdout=subprocess.PIPE).communicate()
+            self._view.processEvents()
+            if re.search("^installed: ", snap_info[0], re.MULTILINE):
+                logging.debug("Snap %s is installed" % snap)
+                # its not tracking the release channel so don't refresh
+                if not re.search(r"^tracking:.*%s" % from_channel,
+                                 snap_info[0], re.MULTILINE):
+                    logging.debug("Snap %s is not tracking the release channel"
+                                  % snap)
+                    continue
+                snap_object['command'] = 'refresh'
+            else:
+                # Do not replace packages not installed
+                cache = self.controller.cache
+                if (deb and (deb not in cache or not cache[deb].is_installed)):
+                    logging.debug("Deb package %s is not installed. Skipping "
+                                  "snap package %s installation" % (deb, snap))
+                    continue
+
+                match = re.search(r"snap-id:\s*(\w*)", snap_info[0])
+                if not match:
+                    logging.debug("Could not parse snap-id for the %s snap"
+                                  % snap)
+                    continue
+                snap_object['command'] = 'install'
+                snap_object['deb'] = deb
+                snap_object['snap-id'] = match[1]
+            snap_object['channel'] = to_channel
+            self._snap_list[snap] = snap_object
+        for snap, (deb, from_channel) in unseeded_snaps.items():
+            snap_object = {}
+            # check to see if the snap is already installed
+            snap_info = subprocess.Popen(["snap", "info", snap],
+                                         universal_newlines=True,
+                                         stdout=subprocess.PIPE).communicate()
+            self._view.processEvents()
+            if re.search("^installed: ", snap_info[0], re.MULTILINE):
+                logging.debug("Snap %s is installed" % snap)
+                # its not tracking the release channel so don't remove
+                if not re.search(r"^tracking:.*%s" % from_channel,
+                                 snap_info[0], re.MULTILINE):
+                    logging.debug("Snap %s is not tracking the release channel"
+                                  % snap)
+                    continue
+
+                snap_object['command'] = 'remove'
+
+                # check if this snap is being used by any other snaps
+                conns = subprocess.Popen(["snap", "connections", snap],
+                                         universal_newlines=True,
+                                         stdout=subprocess.PIPE).communicate()
+                self._view.processEvents()
+
+                for conn in conns[0].split('\n'):
+                    conn_cols = conn.split()
+                    if len(conn_cols) != 4:
+                        continue
+                    plug = conn_cols[1]
+                    slot = conn_cols[2]
+
+                    if slot.startswith(snap + ':'):
+                        plug_snap = plug.split(':')[0]
+                        if plug_snap != '-' and \
+                           plug_snap not in unseeded_snaps:
+                            logging.debug("Snap %s is being used by %s. "
+                                          "Switching it to stable track"
+                                          % (snap, plug_snap))
+                            snap_object['command'] = 'refresh'
+                            snap_object['channel'] = 'stable'
+                            break
+
+                self._snap_list[snap] = snap_object
+        return self._snap_list
